# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import base64
import binascii
import logging

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import (
    Cipher, modes, algorithms
)

import boto3
import botocore
import botocore.exceptions
import libnacl.secret

KEY_ID_FIELD = "key_id"
CIPHERTEXT_FIELD = "ciphertext"
KEY_MANAGER_FIELD = "key_manager"
ALGORITHM_FIELD = "algorithm"

logger = logging.getLogger("biscuit")


def default_kms_factory(region):
    return boto3.client('kms', region_name=region)


def biscuit(kms_client_factory=default_kms_factory):
    """constructs a Biscuit with default settings.

    kms_client_factory is a function that accepts a region string and
    returns a boto3 KMS client."""
    algorithms = {
        'secretbox': SecretBoxAlgo(),
        'aesgcm256': AesGcm256Algo(),
        'none': PlainAlgo(),
    }
    managers = {
        'kms': AwsKmsKeyManager(kms_client_factory),
        'testing': TestingKeyManager()
    }
    return Biscuit(algorithms, managers)


class Biscuit:
    """Biscuit is a reader for encrypted secrets.

    Example:

        from biscuit import biscuit
        secrets = biscuit()
        secrets.read_yaml(open("secrets.yml", "r"))
        decrypted = secrets.get("launch_codes")
    """

    def __init__(self, algorithms, managers):
        """constructor.

        algorithms is a map of {name -> [algorithm, ...]}.
        managers is a map of {name -> [manager, ...]}.
        """
        self.algorithms = algorithms
        self.managers = managers
        self.entries = {}

    def update(self, entries):
        """update updates the internal list of entries with entries from
        the provided map.

        entries is a map of {name -> [value, ...]}.
        """
        self.entries.update(entries)
        return self

    def get(self, name):
        """get decrypts a single entry.

        if there are multiple entries for a name, the first one to
        successfully decrypt will be returned.

        if none of the entries are able to be decrypted, this method returns
        None.
        """
        for entry in self.entries[name]:
            try:
                algo = self.algorithms[entry[ALGORITHM_FIELD]]
                key = None
                if algo.requires_key():
                    key = self.managers[entry[KEY_MANAGER_FIELD]](name, entry)
                plaintext = algo(key, base64.decodestring(
                    entry[CIPHERTEXT_FIELD]))
                return plaintext
            except (ValueError,
                    binascii.Error,
                    botocore.exceptions.ClientError,
                    botocore.exceptions.BotoCoreError) \
                    as e:
                logger.warning("Error decrypting: %s", e)
                continue
        return None


class AwsKmsKeyManager:

    def __init__(self, factory):
        self.factory = factory

    def __call__(self, name, value):
        client = self.factory(region_from_arn(value[KEY_ID_FIELD]))
        decrypt_key_response = client.decrypt(
            CiphertextBlob=base64.decodestring(value["key_ciphertext"]),
            EncryptionContext={'SecretName': name})
        return decrypt_key_response["Plaintext"]


class TestingKeyManager:

    def __call__(self, key, ciphertext):
        return ("x" * 32).encode('ascii')


class SecretBoxAlgo():

    def requires_key(self):
        return True

    def __call__(self, key, ciphertext):
        box = libnacl.secret.SecretBox(key)
        return box.decrypt(ciphertext)


class PlainAlgo():

    def requires_key(self):
        return False

    def __call__(self, _, ciphertext):
        return ciphertext


class AesGcm256Algo():

    def requires_key(self):
        return True

    def __call__(self, key, ciphertext):
        # format is (ciphertext, gcm_tag, nonce)
        nonce = ciphertext[-12:]
        ciphertext_no_nonce = ciphertext[:-12]
        gcm_tag = ciphertext_no_nonce[-16:]
        ciphertext_only = ciphertext_no_nonce[:-16]

        decryptor = Cipher(algorithms.AES(key),
                           modes.GCM(nonce, gcm_tag),
                           backend=default_backend()).decryptor()

        return decryptor.update(ciphertext_only) + decryptor.finalize()


def region_from_arn(arn):
    if arn.startswith("arn:") and arn.split(":") > 6:
        return arn.split(":")[3]
    return None



#!/usr/bin/python
# coding=utf-8

import binascii
import collections
import unittest

import biscuit
import os.path
import yaml

test_vector = collections.namedtuple('test_vector',
                                     ['key', 'nonce', 'plaintext', 'result'])


class TestAesGcm256(unittest.TestCase):
    # Test vectors taken from https://golang.org/src/crypto/cipher/gcm_test.go
    vectors = [
        test_vector(
            '11754cd72aec309bf52f7687212e8957',
            '3c819d9a9bed087615030b65',
            '',
            '250327c674aaf477aef2675748cf6971'
        ),
        test_vector(
            'ca47248ac0b6f8372a97ac43508308ed',
            'ffd2b598feabc9019262d2be',
            '',
            '60d20404af527d248d893ae495707d1a',
        ),
        test_vector(
            '7fddb57453c241d03efbed3ac44e371c',
            'ee283a3fc75575e33efd4887',
            'd5de42b461646c255c87bd2962d3b9a2',
            '2ccda4a5415cb91e135c2a0f78c9b2fdb36d1df9b9d5e596f83e8b7f52971cb3',
        )
    ]

    def test_vectors(self):
        method = biscuit.AesGcm256Algo()

        for v in self.vectors:
            key = binascii.unhexlify(v.key)
            plaintext = binascii.unhexlify(v.plaintext)
            ciphertext = binascii.unhexlify(v.result + v.nonce)
            self.assertEqual(plaintext, method(key, ciphertext))


class TestBiscuit(unittest.TestCase):
    def test_algorithms(self):
        secrets = biscuit.biscuit()
        with open(os.path.join(os.path.dirname(__file__),
                               "hello.yaml"), "r") as fp:
            parsed = yaml.load(fp)
            secrets.update(parsed)
        self.assertEqual("hello", secrets.get("aes"))
        self.assertEqual("hello", secrets.get("secretbox"))
        self.assertEqual("hello", secrets.get("none"))

    def test_region_from_arn(self):
        prefix = "arn:aws:kms:us-west-1:123456789012:"
        tests = {
            prefix + "key/37793df5-ad32-4d06-b19f-bfb95cee4a35": "us-west-1",
            "key/37793df5-ad32-4d06-b19f-bfb95cee4a35": None,
            prefix + "alias/biscuit-x": "us-west-1",
            "alias/biscuit-x": None}
        for arn, region in tests.iteritems():
            self.assertEqual(region, biscuit.region_from_arn(arn))


if __name__ == '__main__':
    unittest.main()



from setuptools import setup


def get_version():
    return open('biscuit/_version.py').readline().strip().strip('"')


setup(
    name='biscuit-py-prima',
    version=get_version(),
    description='biscuit-py decrypts secrets managed by Biscuit.',
    long_description=open('README.rst').read(),
    author='Doug Coker',
    author_email='dcoker@gmail.com',
    url='https://github.com/primait/biscuit-py',
    include_package_data=False,
    license='https://www.apache.org/licenses/LICENSE-2.0',
    packages=[
        'biscuit'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.7'
    ],
    zip_safe=False,
    install_requires=[
        'boto3>=1.3.0,<2',
        'libnacl>=1.4.4,<2',
        'cryptography>=2',
        'pyyaml>=3.11,<5.3'
    ],
    test_suite='tests.biscuit_test',
    entry_points={
        'console_scripts': [
            'biscuitpy=biscuit.__main__:main'
        ]
    },
)
